/*
 * Copyright (c) 2010 CNRS/LAAS
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <math.h>
#include <stdio.h>

#include <portLib.h>
#include <h2mathLib.h>

#include "rmp400Const.h"
#include "server/rmp400Header.h"

static POSTER_ID attitudePosterId = NULL;

static STATUS readAttitudeEstimation(void);


/* rmp400OdometryMethodExec  -  codel EXEC de OdometryMethod
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
rmp400SetOdometryCommandExec(RMP400_ODOMETRY_METHOD *params, int *bilan)
{
	if (params->mode == RMP400_ODO_3D) {
		SDI_F->odometryMode = RMP400_ODO_2D;
		
		if (rmp400POM_SENSOR_POSPosterFind(params->attitudePosterName, 
			&attitudePosterId) == ERROR) {
			*bilan = S_rmp400_CANNOT_READ_POSTER;
			return ETHER;
		}
		
		if (readAttitudeEstimation()) {
			*bilan = S_rmp400_CANNOT_READ_POSTER;
			return ETHER;
		}
		
		SDI_F->odometryMode = RMP400_ODO_3D;
	}
	
	return ETHER;
}


/*
 * Read attitude estimation poster
 */
static STATUS
readAttitudeEstimation(void)
{
	POM_SENSOR_POS full_pos;

	if (rmp400POM_SENSOR_POSPosterRead(attitudePosterId, 
		&full_pos) == ERROR) {
		/* Fall back to ODO 2D */
		printf("rmp400OdometryMethod: Cannot read attitude "
		    "estimation poster\n");
		return ERROR;
	}
	SDI_F->attitude = full_pos.sensorToMain.euler;
	return OK;
}

STATUS
rmp400Odo3d(int *report, double period)
{
	GENPOS_CART_STATE *robot = &SDI_F->robot;
	RMP400_CART_3DSTATE *robot3d = &SDI_F->robot3d;
	RMP400_STATUS_STR *status = &SDI_F->status;
	POM_EULER *attitude = &SDI_F->attitude;
	int i;
	static double mpitch = 0.0, oldPitch = -1.0;
	static int pausePitch = 0;
	static double lastVel = 0.0;
	static double odoPitch = 0.0;
	static double lastLargeAccelTime = 0.0;
	static double history[2*RMP400_ZODOCOR_AVGSIZE] = {0};
	static int history_i = 0;

	/* 3D odometry */
	if (SDI_F->odometryMode == RMP400_ODO_3D) {
		if (readAttitudeEstimation() == ERROR) {
			printf("switching back to odo2D\n");
			SDI_F->odometryMode = RMP400_ODO_2D;
		} else
		{
			/* pause pitch updates if acceleration is too large */
			double accel = (robot->v - lastVel) / period;
			lastVel = robot->v;
			history[history_i++] = attitude->pitch;
			if (history_i >= 2*RMP400_ZODOCOR_AVGSIZE) history_i = 0;
			
			if (fabs(accel) > RMP400_ZODOCOR_ACCTHRES)
			{
				lastLargeAccelTime = 0.0;
				if (!pausePitch)
				{
					/* Compute pitch value to use next during the pitch update pause.
					 * We don't use the latest RMP400_ZODOCOR_AVGSIZE values because they are
					 * often already contaminated by the acceleration (and decreasing the
					 * acceleration threshold would be too sensitive to noise).
					 * We average the RMP400_ZODOCOR_AVGSIZE values before to filter the noise.
					 */
					odoPitch = 0.0;
					for(i = 0; i < RMP400_ZODOCOR_AVGSIZE; ++i)
						odoPitch += history[(history_i+i)%(2*RMP400_ZODOCOR_AVGSIZE)];
					odoPitch /= (2*RMP400_ZODOCOR_AVGSIZE);
					pausePitch = 1;
				}
			} else
			{
				if (pausePitch)
				{
					lastLargeAccelTime += period;
					if (lastLargeAccelTime > RMP400_ZODOCOR_LATENCY)
						pausePitch = 0;
				}
				if (!pausePitch)
					odoPitch = attitude->pitch;
			}
		}
	}
	if (SDI_F->odometryMode == RMP400_ODO_2D) {
		mpitch = 0;
		oldPitch = -1;
		memset(attitude, 0, sizeof(POM_EULER));
	}
	robot3d->pitch = attitude->pitch;
	robot3d->roll = attitude->roll;
	/* XXXX For now keep the position in the Z=0 frame */
	robot3d->theta  = robot->theta;
	robot3d->xRob = robot->xRob;
	robot3d->yRob = robot->yRob;
	robot3d->zRob += - (robot->v * period) * sin(odoPitch);
	robot3d->xRef = robot->xRef;
	robot3d->yRef = robot->yRef;
	robot3d->zRef = 0.0;
	robot3d->v = robot->v;
	robot3d->vt = 0.0;
	robot3d->w = robot->w;

	return OK;
}
